/*
 * Blockbuster 2000
 * ----------------
 *
 * See it in action: https://www.youtube.com/shorts/b4LMFJF6wsM
 *
 * Context: https://lcamtuf.substack.com/p/recreating-classic-games-with-a-bare
 *
 * Be sure to also download font-inline.h with this file. If you're interested
 * in a standalone parallel-mode SSD1309 driver, use ssd1309.c instead.
 *
 * The code isn't award-worthy - I commented it well but didn't refactor
 * clunkiness left over from experimentation. This runs on ATmega1284 with
 * CKDIV8 fuse disabled for 8 MHz operation. JTAGEN fuse also needs to be
 * disabled to allow full use of PORTC inputs.
 *
 * Circuit connections:
 *
 * - PORTA to SSD1309 data bus
 *
 * - PORTB bit 0 to SSD1309 reset line
 *         bit 1 to SSD1309 write select line
 *         bit 2 to SSD1309 command/data select line
 *
 * - PORTC bits 0-2 to up / down / fire buttons, respectively
 *
 * Complaints to: <lcamtuf@coredump.cx>
 *
 *
 *
 * Analog paddle + synth output mod
 * --------------------------------
 *
 * This is a little mod of the original that uses potentiometer on pin PC0 to
 * control the paddle.
 *
 * The OLED display controller was changed to SSD1306 and is connected via I2C.
 *
 * There's no beeper but the game outputs MIDI notes (currently on channel 4)
 * as well as program changes to change instruments (programs 90-93).
 *
 *
 * Pinout:
 *
 * - PC0: paddle position voltage in
 * - PC4: SSD1306 SDA
 * - PC5: SSD1306 SCL
 * - PB2: left button
 * - PB3: right button
 * - PB4: fire button
 * - PD1: MIDI out
 * - PD2: beeper (only if REAL_BEEPER defined)
 *
 * This mod was done by <thement@ibawizard.net>
 */
#include <avr/io.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <util/delay.h>
#include <avr/eeprom.h>
#include <avr/pgmspace.h>

/* User-friendly typedefs */

typedef int8_t   s8;
typedef uint8_t  u8;
typedef int16_t  s16;
typedef uint16_t u16;
typedef int32_t  s32;
typedef uint32_t u32;

/* Display fonts. */

#include "font-inline.h"
#include "defs.h"
#include "i2c.h"
#include "ssd1306.h"
#include "adc.h"
#include "serial.h"

#define C_UP       0
#define C_DOWN     1
#define C_FIRE     2

#define BUTTONS ((PINB) >> 2)

/* Screen update function. The organization of SSD1306 memory is unusual
   in a way that makes screen management rather frustrating. For starters,
   every byte sent to the controller maps to a *vertical* stack of 8
   subsequent pixels. Then, in the default mode, the write pointer advances
   to the right, allowing you to write another 8-pixel stack in the next
   column. Once you reach the last column, you need to manually move the write
   pointer to (0, prev_y + 8).  */

static u8 cur_screen[128 * 8]; /* RAM cost: 1 kB */

static void oled_update(void) {
  ssd1306_send_screen(cur_screen);
}

/* Get the display in a predictable state at boot. */

static void oled_init(void) {
  i2c_init();
  ssd1306_init();
}

/* Macros to manipulate individual pixels in the framebuffer. */

#define set_pixel(_x, _y) do { \
    u8 _tx = (_x), _ty = (_y); \
    cur_screen[((_ty & 63) >> 3) * 128 + (_tx & 127)] |= (1 << (_ty & 7)); \
  } while (0)

#define clear_pixel(_x, _y) do { \
    u8 _tx = (_x), _ty = (_y); \
    cur_screen[((_ty & 63) >> 3) * 128 + (_tx & 127)] &= (1 << (_ty & 7)); \
  } while (0)

#define toggle_pixel(_x, _y) do { \
    u8 _tx = (_x), _ty = (_y); \
    cur_screen[((_ty & 63) >> 3) * 128 + (_tx & 127)] ^= (1 << (_ty & 7)); \
  } while (0)

/* A function to draw a sprite, up to 8 pixel wide, in the framebuffer.

   The pgm_data parameter should point to a PROGMEM sprite. The height of the
   sprite should be specified via ylen. Subsequent bytes of pgm_data will be
   interpreted as subsequent lines of the sprite. If the sprite is fewer than
   8 pixel wide, it's best to pad it on the right.

   The sprite is XORed onto the framebuffer, which simplifies motion: you can
   draw the sprite again at the old position to effectively erase it and
   restore background, and then draw at the new coordinates.

   Note that we read sprite rows MSB to LSB, which corresponds to the intuitive
   left-to-right representation used to design them by hand. */

static void draw_sprite8(u8 x, u8 y, u8 ylen, const u8* pgm_data) {
  u8 sx, sy;

  for (sy = 0; sy < ylen; sy++) {
    u8 d = pgm_read_byte(pgm_data);
    for (sx = 0; sx < 8; sx++)
      if (d & (128 >> sx)) toggle_pixel(x + sx, y);
    pgm_data++;
    y++;
  }
}

/* This is a function similar to draw_sprite8, except it accepts 16-bit sprites. */

static void __attribute__((unused)) draw_sprite16(u8 x, u8 y, u8 ylen,
                                                  const u16* pgm_data) {
  u8 sx, sy;

  for (sy = 0; sy < ylen; sy++) {
    u16 d = pgm_read_word(pgm_data);
    for (sx = 0; sx < 16; sx++)
      if (d & (32768 >> sx)) toggle_pixel(x + sx, y);
    pgm_data++;
    y++;
  }
}

/* This function puts the specified text onto the framebuffer. The function
   doesn't support newlines; all unrecognized characters are output as spaces.

   With a larger charset, you'd want to replace this with a lookup function or
   make a single continuous character map in font-inline.h. That said, for a
   small charset, the current approach conserves memory. */

static void draw_text(u8 x, u8 y, const char* text) {
  while (*text) {
    const u8* ptr = 0;
    switch (*text) {
      case 'A' ... 'Z': ptr = fnt_upper[*text - 'A'];   break;
      case 'a' ... 'z': ptr = fnt_lower[*text - 'a'];   break;
      case '0' ... '9': ptr = fnt_digits[*text - '0'];  break;
      case '.':         ptr = fnt_specials[FNT_PERIOD]; break;
      case ',':         ptr = fnt_specials[FNT_COMMA];  break;
      case '!':         ptr = fnt_specials[FNT_EXMARK]; break;
      case '?':         ptr = fnt_specials[FNT_QMARK];  break;
      case ':':         ptr = fnt_specials[FNT_COLON];  break;
      case '+':         ptr = fnt_specials[FNT_PLUS];   break;
      case '-':         ptr = fnt_specials[FNT_MINUS];  break;
    }
    if (ptr) draw_sprite8(x, y, 6, ptr);
    text++;
    x += 6;
  }
}

/* The following functions draw horizontal / vertical line segments in
   the framebuffer. */

static void draw_hline(u8 x, u8 y, u8 xlen) {
  while (xlen--) toggle_pixel(x + xlen, y);
}

static void draw_vline(u8 x, u8 y, u8 ylen) {
  while (ylen--) toggle_pixel(x, y + ylen);
}

static void draw_box(u8 x, u8 y, u8 xlen, u8 ylen) {
  draw_hline(x, y, xlen);
  draw_vline(x, y + 1, ylen - 1);
  draw_hline(x + 1, y + ylen - 1, xlen - 1);
  draw_vline(x + xlen - 1, y + 1, ylen - 2);
}

/* Functions to clear the entire screen or portion thereof. */

#define clear_screen() \
  do { memset(cur_screen, 0, sizeof(cur_screen)); } while (0)

static void draw_solid_rectangle(u8 x, u8 y, u8 xlen, u8 ylen, u8 color) {
  u8 sx;

  if (color) {
    while (ylen--) for (sx = 0; sx < xlen; sx++) set_pixel(x + sx, y + ylen);
  } else {
    while (ylen--) for (sx = 0; sx < xlen; sx++) clear_pixel(x + sx, y + ylen);
  }
}

/* Generic sound handling routine. */

#ifdef REAL_BEEPER
static void beep(u16 freq, u16 msec) {
  u16 us100 = 5000 / freq;
  u16 cycles = msec * 5 / us100;
  u16 i;

  while (cycles--) {
    PORTD = 4;
    for (i = 0; i < us100; i++) _delay_us(100);
    PORTD = 0;
    for (i = 0; i < us100; i++) _delay_us(100);
  }

}
#else
#define MIDI_CHAN 3

static u8 cur_pgm = 0xff;

/* Send out MIDI program change to switch patch */
static void midi_set_pgm(u8 pgm) {
  if (pgm != cur_pgm) {
    serial_write(0xc0 | MIDI_CHAN);
    serial_write(pgm);
    cur_pgm = pgm;
    /* This delay is mandatory: it takes Monologue some time to switch
     * the patch */
    _delay_ms(100);
  }
}

/* Write a 3-byte midi command */
static void midi_write(u8 a, u8 b, u8 c) {
  serial_write(a | MIDI_CHAN);
  serial_write(b);
  serial_write(c);
}

static void beep(u16 freq, u16 msec) {
  /* Convert frequency to MIDI note. This code does round the calculated
   * note to a nearest semitone because MIDI doesn't support microtunning.
   * (But Monologue *does* support custom tuning tables, so this might
   * be a possible future improvement). */
  u8 midi_note = 43 + (u8)roundf(12.0 * logf((float)freq / 100.0) / logf(2));
  u8 midi_chan = 3;
  /* Note on */
  midi_write(0x90, midi_note, 0x7f);
  while (msec > 0) {
    _delay_ms(1);
    msec--;
  }
  /* Note off */
  midi_write(0x80, midi_note, 0x00);
}
#endif

/* Assorted in-game sound effects. */

enum {
  BLIP_INST,
  SYNTH_INST,
  FLUTE_INST,
  ORGAN_INST,
};

static inline void set_inst(u8 kind) {
  /* The instruments for our game are stored in patches 90..93 */
  byte pgm = 90;
  switch (kind) {
    case BLIP_INST: pgm = 90; break;
    case SYNTH_INST: pgm = 91; break;
    case FLUTE_INST: pgm = 92; break;
    case ORGAN_INST: pgm = 93; break;
  }
  midi_set_pgm(pgm - 1);
}

static void boot_sound(void) {
  set_inst(SYNTH_INST);
  beep(100, 100);
  beep(200, 100);
  beep(400, 100);
  beep(800, 100);
}

static void death_sound(void) {
  set_inst(FLUTE_INST);
  beep(200, 200);
  beep(100, 400);
  beep(80, 600);
  _delay_ms(800);
}

static void game_over_sound(void) {
  set_inst(ORGAN_INST);
  beep(240, 50); beep(220, 50); beep(200, 200); beep(70, 100);
  _delay_ms(50);
  beep(120, 50); beep(110, 50); beep(100, 400); beep(70, 100);
  _delay_ms(50);
  beep(70, 50); beep(60, 50); beep(50, 1400);
}

static void victory_sound(void) {
  set_inst(SYNTH_INST);
  beep(300, 100);
  beep(400, 100);
  beep(500, 100);
  beep(600, 100);
}

static void level_sound(void) {
  set_inst(SYNTH_INST);
  _delay_ms(250); beep(100, 10);
  _delay_ms(250); beep(200, 10);
  _delay_ms(250); beep(300, 10);
  _delay_ms(250); beep(500, 50);
  _delay_ms(900);
}

static void key_sound(void) {
  set_inst(BLIP_INST);
  beep(200, 10);
}

static void splash_click_sound(void) {
  set_inst(BLIP_INST);
  beep(400, 100);
}

static void bounce_sound(u8 combo) {
  set_inst(BLIP_INST);
  beep(800 + combo * 100, 10);
}

/* High score handling. We keep a tally of 6 scores in the EEPROM,
   with several predefined players to choose from. The EEPROM is
   always initialized to 0xFF, so we use that value to indicate
   an empty high score slot. */

#define HS_CHILD01 0x10
#define HS_CHILD02 0x11
#define HS_CHILD03 0x12
#define HS_MOM     0x13
#define HS_DAD     0x14
#define HS_NONE    0xFF

static u8  high_who[6];
static u16 high_score[6];

/* Load high scores from EEPROM. The layout is somewhat arbitrary. */

static void load_high_scores(void) {
  u8 i;
  for (i = 0; i < 6; i++) {
    high_who[i] = eeprom_read_word((const u16*)(0x10 + i * 2));
    if (high_who[i] != HS_NONE)
      high_score[i] = eeprom_read_word((const u16*)(0x40 + i * 2));
  }
}

/* Save high scores to EEPROM. */

static void save_high_scores(void) {
  u8 i;
  for (i = 0; i < 6; i++) {
    eeprom_write_word((u16*)(0x10 + i * 2), high_who[i]);
    eeprom_write_word((u16*)(0x40 + i * 2), high_score[i]);
  }
}

/* Show high score splash screen. It's a bit lazy to use sprintf(), but AVR
   libraries support it and it doesn't take up too much program memory. We
   did manual conversions in dino.c, so... */

static void show_high_scores(void) {
  char tmp[16];
  u8 i;

  key_sound();
  clear_screen();

  draw_solid_rectangle(28, 6, 69, 9, 1);
  draw_text(28, 8, "HIGH SCORES");

  for (i = 0; i < 6; i++) {
    switch (high_who[i]) {
      case HS_CHILD01: sprintf(tmp, "Child01 %05u", high_score[i]); break;
      case HS_CHILD02: sprintf(tmp, "Child02 %05u", high_score[i]); break;
      case HS_CHILD03: sprintf(tmp, "Child03 %05u", high_score[i]); break;
      case HS_MOM:     sprintf(tmp, "    Mom %05u", high_score[i]); break;
      case HS_DAD:     sprintf(tmp, "    Dad %05u", high_score[i]); break;
      default:         sprintf(tmp, "------- %05u", high_score[i]);
    }

    draw_text(20, 23 + i * 6, tmp);

  }

  oled_update();

  /* Wait for key release. */
  while ((BUTTONS & 0b111) != 0b111) _delay_ms(1);
  /* Wait for keypress. */
  while ((BUTTONS & 0b111) == 0b111) _delay_ms(1);
  key_sound();

}

/* Handle high score. This routine allows the user to cycle through predefined
   players, and then updates the working arrays and EEPROM accordingly. */

static void enter_high_score(u16 score) {
  u8 who = HS_CHILD01;
  u8 pc;
  u8 first = 1, pos = 5;

  clear_screen();
  draw_solid_rectangle(28, 6, 68, 9, 1);
  draw_text(28, 8, "HIGH SCORE!");
  draw_text(26, 23, "Select name:");

  /* Loop here when cycling through names. Delete old text and display the
      new choice. */

another_name:

  draw_solid_rectangle(30, 35, 80, 8, 0);

  switch (who) {
    case HS_CHILD01: draw_text(30, 35, "- Child01 -"); break;
    case HS_CHILD02: draw_text(30, 35, "- Child02 -"); break;
    case HS_CHILD03: draw_text(30, 35, "- Child03 -"); break;
    case HS_MOM: draw_text(30, 35,     "- Mom     -"); break;
    case HS_DAD: draw_text(30, 35,     "- Dad     -"); break;
  }

  oled_update();

  /* On the first visit here, make some congratulatory noises. We do it here
     so that there's something already displayed on the screen. */

  if (first) {
    victory_sound();
    victory_sound();
    victory_sound();
    first = 0;
  }

  /* Wait for key release, then keypress. */

  do { pc = ((~BUTTONS) & 0b111); _delay_ms(1); } while (pc);
  do { pc = ((~BUTTONS) & 0b111); _delay_ms(1); } while (!pc);
  key_sound();

  /* Handle up / down scrolling... */

  if (pc & _BV(C_UP)) {
    who--;
     if (who < HS_CHILD01) who = HS_DAD;
     goto another_name;
  }

  if (pc & _BV(C_DOWN)) {
    who++;
    if (who > HS_DAD) who = HS_CHILD01;
    goto another_name;
  }

  /* Alright, if we're here, the player must've pressed "fire". Update
     high scores, display the new line up, and return. */

  bounce_sound(1);

  /* Find insertion point. */

  while (pos && (score > high_score[pos-1])) {
    high_who[pos] = high_who[pos - 1];
    high_score[pos] = high_score[pos - 1];
    pos--;
  }

  high_score[pos] = score;
  high_who[pos] = who;

  save_high_scores();
  show_high_scores();

}

/* Display boot-up splash screen. This allows the user to start a new game or
   view high scores. */

static void splash_screen(void) {
  u16 cycle;
  u8  released;

restart_splash:

  cycle = 0;
  released = 0;

  clear_screen();
  draw_box(0, 0, 128, 64);
  draw_solid_rectangle(16, 18, 99, 9, 1);
  draw_text(16, 20, "BlockBuster 7000");
  draw_text(5, 41, "arrows for hi score");

  while (1) {

    /* Simple text animation. */

    if (!cycle) {
      draw_text(28, 31, "fire to play");
      oled_update();
    }

    cycle = (cycle + 1) % 200;
    _delay_ms(1);

    /* Wait for keypress, but ignore any key held before this loop was
       entered. We do it this way instead of a blocking loop to keep
       the text animation going. After that, either show high scores
       or return to start the actual game. */

    if ((BUTTONS & 0b111) != 0b111) {
      if (!released) continue;
      if (BUTTONS & _BV(C_FIRE)) { show_high_scores(); goto restart_splash; }
      splash_click_sound();
      while ((BUTTONS & 0b111) != 0b111) _delay_ms(1);
      return;
    } else released = 1;

  }

}

/* Paddle and ball sprites. */

static const u8 PROGMEM paddle[7] = {
  0b00100000,
  0b01100000,
  0b10100000,
  0b10100000,
  0b10100000,
  0b01100000,
  0b00100000
};

static const u8 PROGMEM ball[3] = {
  0b01100000,
  0b11110000,
  0b01100000,
};

/* Disappearing ball (game over sequence). */

static const u8 PROGMEM ball_off1[3] = {
  0b11000000,
  0b11100000,
  0b11000000,
};

static const u8 PROGMEM ball_off2[3] = {
  0b10000000,
  0b11000000,
  0b10000000,
};

static const u8 PROGMEM ball_off3[3] = {
  0b00000000,
  0b10000000,
  0b00000000,
};

/* Bricks. The first type takes one hit to destroy'; the second requires two
   hits; and the third requires three hits. */

static const u8 PROGMEM brick_type1[6] = {
  0b11100000,
  0b10100000,
  0b10100000,
  0b10100000,
  0b10100000,
  0b11100000
};

static const u8 PROGMEM brick_type2[6] = {
  0b11100000,
  0b10100000,
  0b11100000,
  0b11100000,
  0b10100000,
  0b11100000
};

static const u8 PROGMEM brick_type3[6] = {
  0b11100000,
  0b11100000,
  0b11100000,
  0b11100000,
  0b11100000,
  0b11100000
};

/* Ball motion constants. These must be pairwise-complementary, because we do
   ^= 0xFF to reverse motion here and there. */

#define BALL_DIR_X_LEFT  0x00
#define BALL_DIR_X_RIGHT 0xFF
#define BALL_DIR_Y_UP    0x00
#define BALL_DIR_Y_DOWN  0xFF

/* Bricks are organized into 7 rows with 6 brick slots each. Brick types:
   0 - no brick; 1 - single-hit brick; 2 - two-hit brick;
   3 - three-hit brick. */

static u8 bricks[7][6];

/* Configure blocks and return game speed per level. */

#define BRICK_LAYOUT(_a, _b, _c, _d, _e, _f) do { \
    bricks[row][0] = (_a); \
    bricks[row][1] = (_b); \
    bricks[row][2] = (_c); \
    bricks[row][3] = (_d); \
    bricks[row][4] = (_e); \
    bricks[row][5] = (_f); \
    bricks_left += !!(_a) + !!(_b) + !!(_c) + !!(_d) + !!(_e) + !!(_f); \
  } while (0)

static u8 bricks_left;

static u8 init_level(u8 level) {
  u8 row;

  bricks_left = 0;

  for (row = 0; row < 7; row++) {

    switch (level) {

      case 1: if (row & 1) BRICK_LAYOUT(1, 0, 1, 0, 1, 0); else BRICK_LAYOUT(0, 1, 0, 1, 0, 1); /* 1 1 1  */
              break;                                                                            /*  1 1 1 */

      case 2: BRICK_LAYOUT(1, 1, 1, 1, 1, 1);                                                   /* 111111 */
              break;                                                                            /* 111111 */

      case 3: if (row & 1) BRICK_LAYOUT(0, 1, 0, 2, 0, 1); else BRICK_LAYOUT(1, 0, 2, 0, 1, 0); /*  1 2 1 */
              break;                                                                            /* 1 2 1  */

      case 4: if (row & 1) BRICK_LAYOUT(0, 2, 0, 1, 0, 2); else BRICK_LAYOUT(0, 1, 0, 2, 0, 1); /* 2 1 2  */
              break;                                                                            /*  1 2 1 */

      case 5: if (row & 1) BRICK_LAYOUT(0, 2, 0, 1, 0, 2); else BRICK_LAYOUT(2, 0, 1, 0, 2, 0); /*  2 1 2 */
              break;                                                                            /* 2 1 2  */

      case 6: if (row & 1) BRICK_LAYOUT(2, 0, 2, 0, 2, 0); else BRICK_LAYOUT(0, 2, 0, 2, 0, 2); /* 2 2 2  */
              break;                                                                            /*  2 2 2 */

      case 7: if (row & 1) BRICK_LAYOUT(1, 2, 1, 2, 1, 2); else BRICK_LAYOUT(2, 1, 2, 1, 2, 1); /* 121212 */
              break;                                                                            /* 212121 */

      case 8: if (row & 1) BRICK_LAYOUT(1, 1, 2, 2, 1, 1); else BRICK_LAYOUT(2, 2, 1, 1, 2, 2); /* 112211 */
               break;                                                                           /* 221122 */

      case 9: BRICK_LAYOUT(2, 2, 2, 2, 2, 2);                                                   /* 222222 */
              break;                                                                            /* 222222 */

      case 10: if (row & 1) BRICK_LAYOUT(2, 0, 1, 0, 2, 0); else BRICK_LAYOUT(0, 1, 0, 3, 0, 1); /* 2 1 2  */
               break;                                                                            /*  1 3 1 */

      case 11: if (row & 1) BRICK_LAYOUT(3, 0, 1, 0, 3, 0); else BRICK_LAYOUT(0, 1, 0, 3, 0, 1); /* 3 1 3  */
               break;                                                                            /*  1 3 1 */

      case 12: if (row & 1) BRICK_LAYOUT(0, 3, 0, 2, 0, 3); else BRICK_LAYOUT(2, 0, 3, 0, 2, 0); /*  3 2 3 */
               break;                                                                            /* 2 3 2  */

      case 13: if (row & 1) BRICK_LAYOUT(3, 0, 2, 0, 3, 0); else BRICK_LAYOUT(0, 3, 0, 2, 0, 3); /* 3 2 3  */
               break;                                                                            /*  3 2 3 */

      case 14: if (row & 1) BRICK_LAYOUT(3, 0, 3, 0, 3, 0); else BRICK_LAYOUT(0, 3, 0, 3, 0, 3); /* 3 3 3  */
               break;                                                                            /*  3 3 3 */

      case 15: if (row & 1) BRICK_LAYOUT(1, 3, 1, 3, 1, 3); else BRICK_LAYOUT(2, 1, 2, 1, 2, 1); /* 131313 */
               break;                                                                            /* 212121 */

      case 16: if (row & 1) BRICK_LAYOUT(3, 1, 3, 1, 3, 1); else BRICK_LAYOUT(1, 3, 1, 3, 1, 3); /* 313131 */
               break;                                                                            /* 131313 */

      case 17: if (row & 1) BRICK_LAYOUT(1, 2, 3, 1, 2, 3); else BRICK_LAYOUT(3, 2, 1, 3, 2, 1); /* 123123 */
               break;                                                                            /* 321321 */

      case 18: if (row & 1) BRICK_LAYOUT(3, 2, 3, 2, 3, 2); else BRICK_LAYOUT(2, 3, 2, 3, 2, 3); /* 323232 */
               break;                                                                            /* 232323 */

      case 19: if (row & 1) BRICK_LAYOUT(3, 3, 2, 2, 3, 3); else BRICK_LAYOUT(2, 2, 3, 3, 2, 2); /* 332233 */
               break;                                                                            /* 223322 */

      default: BRICK_LAYOUT(3, 3, 3, 3, 3, 3);
               break;

    }

  }

  /* Choose speed... */

  if (level < 3) return 30;   /* 1-2   */
  if (level < 5) return 25;   /* 3-4   */
  if (level < 10) return 20;  /* 5-9   */
  if (level < 20) return 15;  /* 10-19 */
  if (level == 20) return 10; /* 20    */
  if (level == 21) return 8;  /* 21    */
  if (level == 22) return 6;  /* 22    */
  if (level == 23) return 5;  /* 23    */
  if (level == 24) return 4;  /* 24    */
  return 4;                   /* 25+   */

}

/* Draw bricks on screen. */

static void draw_bricks(void) {
  u8 row, col;

  for (row = 0; row < 7; row++) {
    for (col = 0; col < 6; col++) {
      switch (bricks[row][col]) {
        case 1: draw_sprite8(65 + 10 * col, 11 + (row * 7), 6, brick_type1); break;
        case 2: draw_sprite8(65 + 10 * col, 11 + (row * 7), 6, brick_type2); break;
        case 3: draw_sprite8(65 + 10 * col, 11 + (row * 7), 6, brick_type3); break;
      }
    }
  }

}

/* Splash level information at the beginning. */

static void level_splash(u8 level) {
  char tmp[10];
  clear_screen();
  sprintf(tmp, "LEVEL %02u", level);
  draw_solid_rectangle(40, 25, 50, 9, 1);
  draw_text(40, 27, tmp);
  oled_update();
  level_sound();
}

/* Splash "game over". */

static void game_over_splash(void) {
  draw_solid_rectangle(30, 27, 66, 13, 0);
  draw_solid_rectangle(32, 29, 62, 9, 1);
  draw_text(32, 31, "GAME OVER!");
  oled_update();
}

/* Variable delay helper function. */

#define delay_ms_var8(_i) do { \
    u8 _v = (_i); \
    while (_v--) _delay_ms(1); \
  } while (0)

/* Update level and score on top of the screen. We call this twice, first
   to XOR-erase the old value, and then to show the new one. Again, a bit lazy
   with sprintf(), but it'll do. */

static void draw_lvl_score(u8 level, u16 score) {
  char tmp[32];
  sprintf(tmp, "LVL %02u          %05u", level, score);
  draw_text(0, 0, tmp);
}

/* Main game loop. */

static void game_loop(void) {

  u16 prev_adc = 0;
  u16 cycle;                        /* Cycle counter for timing            */

  u8  ball_docked;                  /* Ball resting on paddle?             */
  u8  paddle_pos, prev_paddle_pos;  /* Paddle location (y only)            */
  u8  ball_dir_x, ball_dir_y;       /* Ball motion direction               */
  u8  ball_pos_x, old_ball_pos_x,   /* Ball position (start on paddle)     */
      ball_pos_y, old_ball_pos_y;
  u8  inv_speed, inv_speed_y = 5;   /* Animation slowness (inverse speed)  */

  u8  prev_pc = 0;                  /* Previous key input bits             */

  u8  level = 1;                    /* Current level                       */
  u8  lives = 3;                    /* Current lives                       */
  u16 score = 0;                    /* Current score                       */
  u8  combo;                        /* Combo length                        */
  u8  last_down = 0;

  /* This is the entry point for a new level, resetting brick layout and
      adjusting speed. */

new_level:

  inv_speed = init_level(level);
  level_splash(level);

  /* This is the entry point for level retry or the continuation of setting up
     a new one; reset display, redraw all items including bricks, init
     variables, then enter the main game loop. */

level_retry:

  cycle = 0;
  combo = 1;
  ball_docked = 1;
  paddle_pos = prev_paddle_pos = 30;
  ball_dir_x = BALL_DIR_X_RIGHT;
  ball_dir_y = BALL_DIR_Y_UP;
  ball_pos_x = old_ball_pos_x = 3;
  ball_pos_y = old_ball_pos_y = paddle_pos + 2;

  clear_screen();
  draw_box(0, 6, 128, 58);
  draw_vline(0, 7, 57);
  draw_sprite8(0, paddle_pos, 7, paddle);
  draw_sprite8(ball_pos_x, ball_pos_y, 3, ball);
  draw_lvl_score(level, score);

  if (lives) draw_sprite8(60, 1, 3, ball);
  if (lives > 1) draw_sprite8(65, 1, 3, ball);
  if (lives > 2) draw_sprite8(70, 1, 3, ball);

  draw_bricks();
  oled_update();

  /* History of last 4 paddle positions to calculate paddle direction and
   * velocity */
#define N_HIST 4
  s8 paddle_pos_hist[N_HIST];
  u8 hist_ptr = 0;
  while (1) {

    /* PART 1: Handle paddle movement and firing. */

    u8 pc = ((~BUTTONS) & 0b111);

    /* Make key sound if new key pressed or released. */

    if (pc != prev_pc) {
      key_sound();
      prev_pc = pc;
    }

    /* Scan the potentiometer with filtering and scaling to 6-bit value */
    pot_scan(0, &prev_adc, &paddle_pos);
    /* Clamp the paddle position */
    if (paddle_pos < 7)
      paddle_pos = 7;
    if (paddle_pos > (63 - 7))
      paddle_pos = 63 - 7;
    /* If ball is docked, move the ball as well */
    if (ball_docked)
      ball_pos_y = paddle_pos + 2;

    /* Calculate paddle velocity from current position and position 4 ticks ago */
    s8 paddle_dy = ((s8)paddle_pos - (s8)paddle_pos_hist[hist_ptr]) >> 2;
    paddle_pos_hist[hist_ptr] = paddle_pos;
    hist_ptr += 1;
    if (hist_ptr >= N_HIST)
      hist_ptr = 0;

    /* Release docked ball if "fire" pressed. */

    if ((pc & _BV(C_FIRE)) && ball_docked) {

      ball_docked = 0;
      ball_dir_x  = BALL_DIR_X_RIGHT;
      ball_dir_y  = (last_down) ? BALL_DIR_Y_DOWN : BALL_DIR_Y_UP;
      inv_speed_y = 5;

      ball_dir_y = paddle_dy;
      if (paddle_dy > 0) { ball_dir_y = BALL_DIR_Y_DOWN; inv_speed_y = 1; }
      if (paddle_dy < 0)   { ball_dir_y = BALL_DIR_Y_UP; inv_speed_y = 1; }

      bounce_sound(0);

    }

    /* PART 2A: Handle horizontal ball movement. */

    if (!ball_docked) {

      u8 col, row;

      if (ball_dir_x == BALL_DIR_X_RIGHT) ball_pos_x++; else ball_pos_x--;

      /* Bounce upon reaching the rear wall (drawn at x = 127, 4px ball
          width). */

      if (ball_pos_x == 124) {
        ball_dir_x ^= 0xFF;
        ball_pos_x  = 123;
        bounce_sound(combo);
      }

      /* Bounce off paddle at x = 2 (paddle is 3px wide, from x=0 to x=2). We do
         this only if the ball and the paddle are aligned for contact. Paddle
         height is 7px, so it spans from y = paddle_pos to y = paddle_pos + 6.
         Ball height is 3px, from y = ball_pos_y to y = ball_pos_y + 2.

         Technically, the ball is drawn rounded, but we treat it as a rectangle
         for this check, which gives the player a tiny bit of "unfair"
         advantage. */

      if (ball_pos_x == 2 && ball_pos_y + 2 >= paddle_pos &&
          ball_pos_y <= paddle_pos + 6) {

        combo = 1;
        ball_dir_x ^= 0xFF;
        ball_pos_x  = 3;
        bounce_sound(0);

        /* If the paddle is in motion, update drift speed. The speed is
           increased if the paddle is traveling in the same direction as the
           ball, or decreased otherwise. Note that inv_speed_y is the
           *inverse* of speed. */

        if (paddle_dy < 0) {
          if (ball_dir_y == BALL_DIR_Y_UP) inv_speed_y--; else inv_speed_y++;
        } else if (paddle_dy > 0) {
          if (ball_dir_y == BALL_DIR_Y_UP) inv_speed_y++; else inv_speed_y--;
        }

        /* Cap speed to something reasonable. */

        if (inv_speed_y > 8) inv_speed_y = 8;
        if (inv_speed_y < 1) inv_speed_y = 1;

      }

      /* Check for horizontal brick bounces.

         Bricks are placed in six columns starting at x = 65 and spaced 10px
         apart. Brick width is 3px; ball width is 4px.

         Within each column, bricks appear in 7 rows, starting at y = 11 and
         7px apart. Brick height is 6px; ball height is 3px. We check for
         front of ball touching the rear of the brick if moving to the right;
         or vice versa. */

      for (col = 0; col < 6; col++) {

        if ((ball_dir_x == BALL_DIR_X_RIGHT && ball_pos_x + 4 == 65 + 10 * col) ||
            (ball_dir_x == BALL_DIR_X_LEFT && ball_pos_x == 65 + 10 * col + 2)) {

          for (row = 0; row < 7; row++) {

            if (bricks[row][col] && ball_pos_y + 2 >= (11 + 7 * row) &&
                ball_pos_y <= (11 + 7 * row) + 5) {

              /* OK, looks like a hit against an existing brick. Erase the
                 old brick... */

              switch (bricks[row][col]) {
                case 1: draw_sprite8(65 + 10 * col, 11 + (row * 7), 6, brick_type1); break;
                case 2: draw_sprite8(65 + 10 * col, 11 + (row * 7), 6, brick_type2); break;
                case 3: draw_sprite8(65 + 10 * col, 11 + (row * 7), 6, brick_type3); break;
              }

              bricks[row][col]--;

              /* And draw a new one, if appropriate.*/

              switch (bricks[row][col]) {
                case 0: bricks_left--; break;
                case 1: draw_sprite8(65 + 10 * col, 11 + (row * 7), 6, brick_type1); break;
                case 2: draw_sprite8(65 + 10 * col, 11 + (row * 7), 6, brick_type2); break;
                case 3: draw_sprite8(65 + 10 * col, 11 + (row * 7), 6, brick_type3); break;
              }

              /* Erase old score, then draw new value. */

              draw_lvl_score(level, score);
              score += combo;
              draw_lvl_score(level, score);
              if (combo < 3) combo++;

              /* Bounce the ball. */

              ball_dir_x ^= 0xFF;
              bounce_sound(combo);

            }
          }
        }
      }

      /* PART 2B: Handle vertical ball movement. */

      if (!(cycle % inv_speed_y) && (cycle % 19)) {

        if (ball_dir_y == BALL_DIR_Y_DOWN) ball_pos_y++; else ball_pos_y--;

        /* Bounce upon hitting the top wall. */

        if (ball_pos_y == 6) {
          ball_dir_y ^= 0xFF;
          ball_pos_y  = 7;
          bounce_sound(combo);
        }

        /* Boounce upon hitting the bottom wall (drawn at y = 63, 3px ball height). */

        if (ball_pos_y  == 61) {
          ball_dir_y ^= 0xFF;
          ball_pos_y  = 60;
          bounce_sound(combo);
        }

      /* Check for vertical brick bounces. This is a bit of a copy-and-paste from above,
         I should abstract out the brick & score update code. */

      for (row = 0; row < 7; row++) {

        if ((ball_dir_y == BALL_DIR_Y_DOWN && ball_pos_y + 2 == 11 + 7 * row) ||
            (ball_dir_y == BALL_DIR_Y_UP && ball_pos_y == 11 + 7 * row + 5)) {

          for (col = 0; col < 6; col++) {

            if (bricks[row][col] && ball_pos_x + 3 >= (65 + 10 * col) &&
                ball_pos_x <= (65 + 10 * col) + 2) {

              /* OK, looks like a hit against an existing brick. Erase the old brick... */

              switch (bricks[row][col]) {
                case 1: draw_sprite8(65 + 10 * col, 11 + (row * 7), 6, brick_type1); break;
                case 2: draw_sprite8(65 + 10 * col, 11 + (row * 7), 6, brick_type2); break;
                case 3: draw_sprite8(65 + 10 * col, 11 + (row * 7), 6, brick_type3); break;
              }

              bricks[row][col]--;

              /* And draw a new one, if appropriate.*/

              switch (bricks[row][col]) {
                case 0: bricks_left--; break;
                case 1: draw_sprite8(65 + 10 * col, 11 + (row * 7), 6, brick_type1); break;
                case 2: draw_sprite8(65 + 10 * col, 11 + (row * 7), 6, brick_type2); break;
                case 3: draw_sprite8(65 + 10 * col, 11 + (row * 7), 6, brick_type3); break;
              }

              /* Erase old score, then draw new value. */

              draw_lvl_score(level, score);
              score += combo;
              draw_lvl_score(level, score);
              if (combo < 3) combo++;

              /* Bounce the ball. */

              ball_dir_y ^= 0xFF;
              bounce_sound(combo);

            }
          }
        }
      }

      }

    }

    /* PART 3: Redraw ball and paddle, update screen, wait. Sprites are drawn
       with XOR, so erasing the old one boils down to re-drawing it at its
       previous x and y. */

    if (ball_pos_x != old_ball_pos_x || ball_pos_y != old_ball_pos_y) {
      draw_sprite8(ball_pos_x, ball_pos_y, 3, ball);
      draw_sprite8(old_ball_pos_x, old_ball_pos_y, 3, ball);
      old_ball_pos_x = ball_pos_x;
      old_ball_pos_y = ball_pos_y;
    }

    if (paddle_pos != prev_paddle_pos) {
      draw_sprite8(0, prev_paddle_pos, 7, paddle);
      draw_sprite8(0, paddle_pos, 7, paddle);
      prev_paddle_pos = paddle_pos;
    }

    oled_update();

    /* A convenient feature to speed up ball travel with the fire button. */

    if ((pc & _BV(C_FIRE)) && inv_speed > 5) _delay_ms(1);
      else delay_ms_var8(inv_speed - 4);

    cycle = (cycle + 1) % 1000;

    /* PART 4: Check for game over condition (paddle miss). We know the
       situation is unrecoverable by the time paddle_pos_x reaches 2, but we
       let it play out to x = 0. At that point, the ball is touching the left
       edge of the screen, and it looks clunky to stop the animation there.
       So, we manually animate the ball disappearing off the screen. */

    /* TODO: Maybe adjust y along the way to make the animation better. Also,
       probably best to move this to a separate function. */

    if (ball_pos_x == 0) {
      delay_ms_var8(inv_speed);
      draw_sprite8(0, ball_pos_y, 3, ball);
      draw_sprite8(0, ball_pos_y, 3, ball_off1); /* 1st column missing  */
      oled_update();
      delay_ms_var8(inv_speed);
      draw_sprite8(0, ball_pos_y, 3, ball_off1);
      draw_sprite8(0, ball_pos_y, 3, ball_off2); /* +2nd column missing */
      oled_update();
      delay_ms_var8(inv_speed);
      draw_sprite8(0, ball_pos_y, 3, ball_off2);
      draw_sprite8(0, ball_pos_y, 3, ball_off3); /* +3rd column missing */
      oled_update();
      delay_ms_var8(inv_speed);
      draw_sprite8(0, ball_pos_y, 3, ball_off3); /* ...all gone!        */
      oled_update();

      if (lives) {
        death_sound();
        _delay_ms(500);
        lives--;
        goto level_retry;
      }

      /* If we're here, we're out of lives. Go through the motions, and
         check if this warrants a high score screen. */

      game_over_sound();
      game_over_splash();
      _delay_ms(700);

      if (score > high_score[5]) enter_high_score(score);

      return;

    }

    /* Win condition. */

    if (!bricks_left) {
      victory_sound();
      level++;
      lives = 3;
      _delay_ms(500);
      goto new_level;
    }

  }

}

/* Main entry point */

int main(void) {

  /* Set up I/O ports */

  DDRB = 0b00000000; /* 2: button up; 3: button down; 4: start */
  PORTB = 0b00011100; /* Button pull-up resistors */
#ifdef REAL_BEEPER
  DDRD = 0b00000100;
#endif

  adc_init();
  /* MIDI out */
  serial_start(31250);
  load_high_scores();
  oled_init();
  boot_sound();

  while (1) {
    splash_screen();
    game_loop();
  }
}
